/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Jianfeng Gao ( jianfeng dot gao at kit dot edu )
 * @date       2022
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/interface/serialization/Eigen.h>
#include <ArmarXCore/util/CPPUtility/TripleBuffer.h>
#include "MP.h"


namespace armarx::control::common::mp
{

    struct TSMPInput : virtual MPInput
    {
        Eigen::Matrix4f pose = Eigen::Matrix4f::Identity();
        Eigen::Vector6f vel = Eigen::Vector6f::Zero();
        double deltaT;
    };

    struct TSMPOutput : virtual MPOutput
    {
        Eigen::Matrix4f pose = Eigen::Matrix4f::Identity();
        Eigen::Vector6f vel = Eigen::Vector6f::Zero();
    };

    using TSMPInputPtr  = std::shared_ptr<TSMPInput>;
    using TSMPOutputPtr = std::shared_ptr<TSMPOutput>;



    class TSMP : virtual public MP
    {
    public:
        TSMP(const MPConfig& c);
        ~TSMP(){};

        void run(MPInputPtr, MPOutputPtr) override;

        Eigen::Quaterniond previousAngularVelocity = Eigen::Quaterniond::Identity();
    };

    typedef std::shared_ptr<TSMP> TSMPPtr;

} /// namespace armarx::control::common::mp

