/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Jianfeng Gao ( jianfeng dot gao at kit dot edu )
 * @date       2022
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <SimoxUtility/meta/enum/EnumNames.hpp>

namespace armarx::control::common
{

    enum class ControllerType
    {
        TSAdmittance,
        TSImpedance,
        KptAdmittance,
        KptImpedance,
        TSAdmittanceMP,
        TSImpedanceMP,
        KptAdmittanceMP,
        KptImpedanceMP,
        KVILImpedanceMP,
        QPWholeBodyImpedance,
        QPWholeBodyVelocity,
    };

    inline const simox::meta::EnumNames<ControllerType> ControllerTypeShort{
            {ControllerType::TSAdmittance, "TSAdmittance"},
            {ControllerType::TSImpedance, "TSImpedance"},
            {ControllerType::KptAdmittance, "KeypointAdmittance"},
            {ControllerType::KptImpedance, "KeypointImpedance"},
            {ControllerType::TSAdmittanceMP, "TSAdmittanceMP"},
            {ControllerType::TSImpedanceMP, "TSImpedanceMP"},
            {ControllerType::KptAdmittanceMP, "KeypointAdmittanceMP"},
            {ControllerType::KptImpedanceMP, "KeypointImpedanceMP"},
            {ControllerType::KVILImpedanceMP, "KVILImpedanceMP"},
            {ControllerType::QPWholeBodyImpedance, "QPWholeBodyImpedance"},
            {ControllerType::QPWholeBodyVelocity, "QPWholeBodyVelocity"}};


    inline const simox::meta::EnumNames<ControllerType> ControllerTypeNames{
        {ControllerType::TSAdmittance, "NJointTaskspaceAdmittanceController"},
        {ControllerType::TSImpedance, "NJointTaskspaceImpedanceController"},
        {ControllerType::KptAdmittance, "NJointKeypointsAdmittanceController"},
        {ControllerType::KptImpedance, "KeypointImpedance"},
        {ControllerType::TSAdmittanceMP, "NJointTSAdmittanceMPController"},
        {ControllerType::TSImpedanceMP, "NJointTSImpedanceMPController"},
        {ControllerType::KptAdmittanceMP, "NJointKeypointsAdmittanceMPController"},
        {ControllerType::KptImpedanceMP, "NJointKeypointsImpedanceMPController"},
        {ControllerType::KVILImpedanceMP, "NJointKVILImpedanceMPController"}};

} // namespace armarx::control::common
